import json


def json_encoder(content, *args, **kwargs):
    """
    Json encoder to be used by service_client. For the moment is just a proxy to json.dumps
    """
    return json.dumps(content)


def json_decoder(content, *args, **kwargs):
    """
    Json decoder parser to be used by service_client
    """
    if not content:
        return None
    json_value = content.decode()
    return json.loads(json_value)



import logging
from asyncio import coroutine, get_event_loop
from urllib.parse import urlparse, urlunsplit
from aiohttp.client import ClientSession
from aiohttp.connector import TCPConnector


class SessionWrapper:

    def __init__(self, session):
        self.set_custom_session(session)

    def __getattr__(self, item):  # pragma: no cover
        return getattr(self._session, item)

    def __setattr__(self, key, value):  # pragma: no cover
        return setattr(self._session, key, value)

    def __str__(self):  # pragma: no cover
        return str(self._session)

    def __repr__(self):  # pragma: no cover
        return repr(self._session)

    def __eq__(self, other):
        return self._session.__eq__(other)

    def set_attr_wrap(self, key, value):
        super(SessionWrapper, self).__setattr__(key, value)

    def set_custom_session(self, session):
        super(SessionWrapper, self).__setattr__('_session', session)


class ServiceClient:

    def __init__(self, name='GenericService', spec=None, plugins=None, config=None,
                 parser=None, serializer=None, base_path='', loop=None, logger=None):
        self._plugins = []

        self.logger = logger or logging.getLogger('serviceClient.{}'.format(name))
        self.name = name
        self.spec = spec or {}
        self.add_plugins(plugins or [])
        self.config = config or {}
        self.parser = parser or (lambda x, *args, **kwargs: x)
        self.serializer = serializer or (lambda x, *args, **kwargs: x)
        self.base_path = base_path
        self.loop = loop or get_event_loop()

        self.connector = TCPConnector(loop=self.loop, **self.config.get('connector', {}))
        self.session = ClientSession(connector=self.connector, loop=self.loop)

    @coroutine
    def call(self, endpoint, payload=None, **kwargs):
        self.logger.debug("Calling service {0}...".format(endpoint))
        endpoint_desc = self.spec[endpoint].copy()
        endpoint_desc['endpoint'] = endpoint

        request_params = kwargs
        session = yield from self.prepare_session(endpoint_desc, request_params)

        request_params['url'] = yield from self.generate_path(endpoint_desc, session, request_params)
        request_params['method'] = endpoint_desc.get('method', 'GET').upper()

        yield from self.prepare_request_params(endpoint_desc, session, request_params)

        self.logger.info("Calling service {0} using {1} {2}".format(endpoint,
                                                                    request_params['method'],
                                                                    request_params['url']))

        payload = yield from self.prepare_payload(endpoint_desc, session, request_params, payload)
        try:
            if request_params['method'] not in ['GET', 'DELETE']:
                try:
                    stream_request = endpoint_desc['stream_request']
                except KeyError:
                    stream_request = False
                if payload and not stream_request:
                    request_params['data'] = self.serializer(payload, session=session,
                                                             endpoint_desc=endpoint_desc,
                                                             request_params=request_params)

            yield from self.before_request(endpoint_desc, session, request_params)

            response = yield from session.request(**request_params)
        except Exception as e:
            self.logger.warn("Exception calling service {0}: {1}".format(endpoint, e))
            yield from self.on_exception(endpoint_desc, session, request_params, e)
            raise e

        yield from self.on_response(endpoint_desc, session, request_params, response)

        try:
            if endpoint_desc['stream_response']:
                return response
        except KeyError:
            pass

        try:
            self.logger.info("Parsing response from {0}...".format(endpoint))
            response.data = self.parser((yield from response.read()),
                                        session=session,
                                        endpoint_desc=endpoint_desc,
                                        response=response)
            yield from self.on_parsed_response(endpoint_desc, session, request_params, response)
        except Exception as e:
            self.logger.warn("[Response code: {0}] Exception parsing response from service "
                             "{1}: {2}".format(response.status, endpoint, e))
            yield from self.on_parse_exception(endpoint_desc, session, request_params, response, e)
            e.response = response
            raise e

        return response

    @coroutine
    def prepare_session(self, endpoint_desc, request_params):
        session = SessionWrapper(self.session)
        yield from self._execute_plugin_hooks('prepare_session', endpoint_desc=endpoint_desc, session=session,
                                              request_params=request_params)
        return session

    @coroutine
    def generate_path(self, endpoint_desc, session, request_params):
        path = endpoint_desc.get('path', '')
        url = list(urlparse(self.base_path))
        url[2] = '/'.join([url[2].rstrip('/'), path.lstrip('/')])
        url.pop()
        path = urlunsplit(url)
        hooks = [getattr(plugin, 'prepare_path') for plugin in self._plugins
                 if hasattr(plugin, 'prepare_path')]
        self.logger.debug("Calling {0} plugin hooks...".format('prepare_path'))
        for func in hooks:
            path = yield from func(endpoint_desc=endpoint_desc, session=session,
                                   request_params=request_params, path=path)

        return path

    @coroutine
    def prepare_request_params(self, endpoint_desc, session, request_params):
        yield from self._execute_plugin_hooks('prepare_request_params', endpoint_desc=endpoint_desc,
                                              session=session, request_params=request_params)

    @coroutine
    def prepare_payload(self, endpoint_desc, session, request_params, payload):
        hooks = [getattr(plugin, 'prepare_payload') for plugin in self._plugins
                 if hasattr(plugin, 'prepare_payload')]
        self.logger.debug("Calling {0} plugin hooks...".format('prepare_payload'))
        for func in hooks:
            payload = yield from func(endpoint_desc=endpoint_desc, session=session,
                                      request_params=request_params, payload=payload)
        return payload

    @coroutine
    def before_request(self, endpoint_desc, session, request_params):
        yield from self._execute_plugin_hooks('before_request', endpoint_desc=endpoint_desc,
                                              session=session, request_params=request_params)

    @coroutine
    def on_exception(self, endpoint_desc, session, request_params, ex):
        yield from self._execute_plugin_hooks('on_exception', endpoint_desc=endpoint_desc,
                                              session=session, request_params=request_params, ex=ex)

    @coroutine
    def on_response(self, endpoint_desc, session, request_params, response):
        yield from self._execute_plugin_hooks('on_response', endpoint_desc=endpoint_desc,
                                              session=session, request_params=request_params, response=response)

    @coroutine
    def on_parse_exception(self, endpoint_desc, session, request_params, response, ex):
        yield from self._execute_plugin_hooks('on_parse_exception', endpoint_desc=endpoint_desc,
                                              session=session, request_params=request_params, response=response, ex=ex)

    @coroutine
    def on_parsed_response(self, endpoint_desc, session, request_params, response):
        yield from self._execute_plugin_hooks('on_parsed_response', endpoint_desc=endpoint_desc, session=session,
                                              request_params=request_params, response=response)

    @coroutine
    def _execute_plugin_hooks(self, hook, *args, **kwargs):
        hooks = [getattr(plugin, hook) for plugin in self._plugins if hasattr(plugin, hook)]
        self.logger.debug("Calling {0} plugin hooks...".format(hook))
        for func in hooks:
            yield from func(*args, **kwargs)

    def add_plugins(self, plugins):
        self._plugins.extend(plugins)

        hook = 'assign_service_client'
        hooks = [getattr(plugin, hook) for plugin in self._plugins if hasattr(plugin, hook)]
        self.logger.debug("Calling {0} plugin hooks...".format(hook))
        for func in hooks:
            func(service_client=self)

    def __getattr__(self, item):

        @coroutine
        def wrap(*args, **kwargs):

            return self.call(item, *args, **kwargs)

        return wrap

    def __del__(self):  # pragma: no cover
        self.session.close()



import string
from string import Formatter

import random


class IncompleteFormatter(Formatter):

    """
    String formatter to safe replace every placeholder. When the placeholder is not
    replaced it remains the same in the string.
    """

    def __init__(self):
        self._substituted_fields = []
        self._not_substituted_fields = []

    def _manage_substituted_field(self, field_name, args, kwargs):
        self._substituted_fields.append(field_name)

    def _manage_not_substituted_field(self, field_name, args, kwargs):
        self._not_substituted_fields.append(field_name)

    def get_field(self, field_name, args, kwargs):
        try:
            val = super(IncompleteFormatter, self).get_field(field_name, args, kwargs)
        except (KeyError, IndexError):
            self._manage_not_substituted_field(field_name, args, kwargs)
            val = '{{{0}}}'.format(field_name), field_name
        else:
            self._manage_substituted_field(field_name, args, kwargs)
        return val

    def get_substituted_fields(self):
        return self._substituted_fields

    def get_not_substituted_fields(self):
        return self._not_substituted_fields


def random_token(length=10):
    """
    Builds a random string.

    :param length: Token length. **Default:** 10
    :type length: int
    :return: str
    """
    return ''.join(random.choice(string.ascii_uppercase + string.digits)
                   for _ in range(length))



from aiohttp.client_reqrep import ClientResponse
from aiohttp.multidict import CIMultiDict
from functools import wraps
from asyncio import coroutine, get_event_loop
from dirty_loader import LoaderNamespaceReversedCached
from .plugins import BasePlugin


class NoMock(Exception):
    pass


class BaseMockDefinition:

    def __init__(self, mock_manager, service_name=None, endpoint=None, offset=0, limit=1):
        self.mock_manager = mock_manager
        self.service_name = service_name
        self.endpoint = endpoint
        self.offset = offset if offset >= 0 else 0
        self.limit = limit if limit >= 0 else 1

    def __call__(self, func):
        @wraps(func)
        @coroutine
        def inner(*args, **kwargs):
            with self:
                yield from func(*args, **kwargs)

        return inner

    def __enter__(self):
        self.mock_manager.push(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.mock_manager.pop(self)


class PatchMockDescDefinition(BaseMockDefinition):

    def __init__(self, patch, *args, **kwargs):
        super(PatchMockDescDefinition, self).__init__(*args, **kwargs)
        self.patch = patch


class UseMockDefinition(BaseMockDefinition):

    def __init__(self, mock, *args, **kwargs):
        super(UseMockDefinition, self).__init__(*args, **kwargs)
        self.mock = mock


class MockManager:

    def __init__(self):
        self.mocks = []

    def patch_mock_desc(self, patch, *args, **kwarg):
        """
        Context manager or decorator in order to patch a mock definition of service
        endpoint in a test.

        :param patch: Dictionary in order to update endpoint's mock definition
        :type patch: dict
        :param service_name: Name of service where you want to use mock. If None it will be used
            as soon as possible.
        :type service_name: str
        :param endpoint: Endpoint where you want to use mock. If None it will be used
            as soon as possible.
        :type endpoint: str
        :param offset: Times it must be ignored before use. Default 0. Only positive integers.
        :type offset: int
        :param limit: Times it could be used. Default 1. 0 means no limit. Only positive integers.
        :type limit: int
        :return: PatchMockDescDefinition
        """

        return PatchMockDescDefinition(patch, self, *args, **kwarg)

    def use_mock(self, mock, *args, **kwarg):
        """
        Context manager or decorator in order to use a coroutine as mock of service
        endpoint in a test.

        :param mock: Coroutine to use as mock. It should behave like :meth:`~ClientSession.request`.
        :type mock: coroutine
        :param service_name: Name of service where you want to use mock. If None it will be used
            as soon as possible.
        :type service_name: str
        :param endpoint: Endpoint where you want to use mock. If None it will be used
            as soon as possible.
        :type endpoint: str
        :param offset: Times it must be ignored before use. Default 0. Only positive integers.
        :type offset: int
        :param limit: Times it could be used. Default 1. 0 means no limit. Only positive integers.
        :type limit: int
        :return: UseMockDefinition
        """
        return UseMockDefinition(mock, self, *args, **kwarg)

    def push(self, mock_description):
        self.mocks.insert(0, mock_description)

    def pop(self, mock_description):
        try:
            self.mocks.remove(mock_description)
        except ValueError:  # pragma: no cover
            pass

    def next_mock(self, service_name, endpoint):
        for i in range(len(self.mocks)):
            candidate = self.mocks[i]
            if (candidate.service_name is not None and candidate.service_name != service_name) \
                    or (candidate.endpoint is not None and candidate.endpoint != endpoint):
                continue

            if candidate.offset > 0:
                candidate.offset -= 1
                continue

            if candidate.limit > 1:
                candidate.limit -= 1
            else:
                self.mocks.pop(i)

            return candidate

        raise NoMock()


mock_manager = MockManager()


class Mock(BasePlugin):

    def __init__(self, namespaces=None):

        self.loader = LoaderNamespaceReversedCached()
        self.loader.register_namespace('default', __name__)

        if namespaces:
            for n, m in namespaces.items():
                self.loader.register_namespace(n, m)

    def _create_mock(self, endpoint_desc, session, request_params, mock_desc, loop):
        """
        The class imported should have the __call__ function defined to be an object directly callable
        """
        try:
            mock_def = mock_manager.next_mock(self.service_client.name,
                                              endpoint_desc['endpoint'])

            if isinstance(mock_def, PatchMockDescDefinition):
                mock_desc.update(mock_def.patch)
            else:
                return mock_def.mock
        except NoMock:
            pass

        return self.loader.factory(mock_desc.get('mock_type'),
                                   endpoint_desc, session,
                                   request_params, mock_desc,
                                   loop=loop)

    @coroutine
    def prepare_session(self, endpoint_desc, session, request_params):
        mock_desc = endpoint_desc.get('mock', {})
        session.set_attr_wrap('request', self._create_mock(endpoint_desc,
                                                           session,
                                                           request_params,
                                                           mock_desc.copy(),
                                                           loop=self.service_client.loop))

        try:
            session.request.set_request_params(request_params)
        except AttributeError:
            pass


class BaseMock:

    def __init__(self, endpoint_desc, session, request_params,
                 mock_desc, loop=None):
        self.endpoint_desc = endpoint_desc
        self.session = session
        self.request_params = request_params
        self.mock_desc = mock_desc
        self.loop = loop or get_event_loop()

    @coroutine
    def __call__(self, *args, **kwargs):
        args = list(args)
        try:
            method = kwargs['method']
        except KeyError:
            method = args.pop()

        try:
            url = kwargs['url']
        except KeyError:
            url = args.pop()

        self.method = method
        self.url = url
        self.args = args
        self.kwargs = kwargs
        self.response = ClientResponse(method, url)
        self.response._post_init(self.loop)
        self.response.status = self.mock_desc.get('status', 200)
        self.response.headers = CIMultiDict(self.mock_desc.get('headers', {}))

        yield from self.prepare_response()

        return self.response


class BaseFileMock(BaseMock):

    @coroutine
    def prepare_response(self):
        filename = self.mock_desc['file']
        self.response._content = self.load_file(filename)


class RawFileMock(BaseFileMock):

    def load_file(self, filename):
        return open(filename, "rb").read()



from setuptools import setup
import os

setup(
    name='aio-service-client',
    url='https://github.com/alfred82santa/aio-service-client',
    author='alfred82santa',
    version='0.2.0',
    license='LGPLv3',
    author_email='alfred82santa@gmail.com',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Development Status :: 4 - Beta'],
    packages=['service_client'],
    include_package_data=True,
    install_requires=['dirty-loader', 'aiohttp'],
    description="Service Client Framework powered by Python asyncio.",
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.rst')).read(),
    test_suite="nose.collector",
    tests_require="nose",
    zip_safe=True,
)



import weakref
import logging
from asyncio import coroutine
from urllib.parse import quote_plus
from aiohttp.helpers import Timeout as TimeoutContext
from aiohttp.multidict import CIMultiDict
from service_client.utils import IncompleteFormatter, random_token


class BasePlugin:

    def assign_service_client(self, service_client):
        self.service_client = service_client

    @property
    def service_client(self):
        return self._service_client()

    @service_client.setter
    def service_client(self, service_client):
        self._service_client = weakref.ref(service_client)


class PathTokens(BasePlugin):

    def __init__(self, default_tokens=None):
        self.default_token = default_tokens or {}

    @coroutine
    def prepare_path(self, endpoint_desc, session, request_params, path):
        tokens = self.default_token.copy()
        try:
            tokens.update(endpoint_desc['path_tokens'])
        except KeyError:
            pass
        tokens.update(request_params)
        formatter = IncompleteFormatter()
        ret = formatter.format(path, **{k: quote_plus(str(v)) for k, v in tokens.items()})
        for field in formatter.get_substituted_fields():
            try:
                request_params.pop(field)
            except KeyError:
                pass
        return ret


class Headers(BasePlugin):

    def __init__(self, default_headers=None):
        self.default_headers = default_headers.copy() if default_headers else {}

    @coroutine
    def prepare_request_params(self, endpoint_desc, session, request_params):
        headers = request_params.get('headers', CIMultiDict()).copy()
        if not isinstance(headers, CIMultiDict):
            headers = CIMultiDict(headers)
        headers.update(self.default_headers)
        headers.update(endpoint_desc.get('headers', {}))
        headers.update(request_params.get('headers', {}))
        request_params['headers'] = headers


class Timeout(BasePlugin):

    def __init__(self, default_timeout=None):
        self.default_timeout = default_timeout

    @coroutine
    def before_request(self, endpoint_desc, session, request_params):
        try:
            timeout = request_params.pop('timeout')
        except KeyError:
            timeout = endpoint_desc.get('timeout', self.default_timeout)

        if timeout is None:
            return

        func = session.request

        @coroutine
        def request_wrapper(*args, **kwargs):
            with TimeoutContext(timeout=timeout):
                yield from func(*args, **kwargs)

        session.set_attr_wrap('request', request_wrapper)


class Elapsed(BasePlugin):

    @coroutine
    def prepare_session(self, endpoint_desc, session, request_params):

        func = session.request

        @coroutine
        def request_wrapper(*args, **kwargs):
            from datetime import datetime
            t = datetime.now()
            response = yield from func(*args, **kwargs)
            response.elapsed = datetime.now() - t
            return response

        session.set_attr_wrap('request', request_wrapper)


class QueryParams(BasePlugin):

    @coroutine
    def prepare_request_params(self, endpoint_desc, session, request_params):
        query_params = endpoint_desc.get('query_params', {}).copy()
        query_params.update(request_params.get('params', {}))
        request_params['params'] = query_params


class BaseLogger(BasePlugin):

    def __init__(self, logger, max_body_length=0, level=logging.INFO,
                 on_exception_level=logging.CRITICAL,
                 on_parse_exception_level=logging.CRITICAL):
        self.logger = logger
        self.max_body_length = max_body_length
        self.level = level
        self.on_exception_level = on_exception_level
        self.on_parse_exception_level = on_parse_exception_level

    def _prepare_body(self, body_string):
        return body_string[:self.max_body_length]

    def _prepare_record(self, endpoint_desc, session, request_params):
        log_data = {'endpoint': endpoint_desc['endpoint'],
                    'service_name': self.service_client.name,
                    'tracking_token': session.tracking_token}

        log_data.update(request_params)

        return log_data

    @coroutine
    def _prepare_request_log_record(self, endpoint_desc, session, request_params):
        try:
            tracking_token = request_params.pop('tracking_token')
        except KeyError:
            tracking_token = random_token()

        session.tracking_token = tracking_token

        log_data = self._prepare_record(endpoint_desc, session, request_params)
        log_data['action'] = 'REQUEST'
        return log_data

    @coroutine
    def _prepare_response_log_record(self, endpoint_desc, session, request_params, response):
        log_data = self._prepare_record(endpoint_desc, session, request_params)
        log_data['action'] = 'RESPONSE'
        log_data['status_code'] = response.status
        log_data['headers'] = response.headers

        try:
            log_data['elapsed'] = response.elapsed
        except AttributeError:  # pragma: no cover
            pass

        try:
            log_data['body'] = self._prepare_body(response.data)
        except AttributeError:
            log_data['body'] = self._prepare_body((yield from response.text()))

        return log_data

    def _prepare_exception(self, log_data, ex):
        log_data['action'] = 'EXCEPTION'
        log_data['exception'] = ex

    @coroutine
    def _prepare_parse_response_exception_log_record(self, endpoint_desc, session, request_params, response, ex):
        log_data = yield from self._prepare_response_log_record(endpoint_desc, session, request_params, response)
        self._prepare_exception(log_data, ex)

        return log_data

    @coroutine
    def _prepare_exception_log_record(self, endpoint_desc, session, request_params, ex):
        log_data = self._prepare_record(endpoint_desc, session, request_params)
        self._prepare_exception(log_data, ex)

        return log_data

    @coroutine
    def on_exception(self, endpoint_desc, session, request_params, ex):
        log_data = yield from self._prepare_exception_log_record(endpoint_desc, session, request_params, ex)
        self.logger.log(self.on_exception_level, str(ex), extra=log_data)

    @coroutine
    def on_parse_exception(self, endpoint_desc, session, request_params, response, ex):
        log_data = yield from self._prepare_parse_response_exception_log_record(endpoint_desc, session,
                                                                                request_params, response, ex)
        self.logger.log(self.on_parse_exception_level, str(ex), extra=log_data)


class InnerLogger(BaseLogger):

    @coroutine
    def _prepare_on_request_log_record(self, endpoint_desc, session, request_params):
        log_data = yield from self._prepare_request_log_record(endpoint_desc, session, request_params)
        try:
            log_data['body'] = self._prepare_body(log_data['data'])
            del log_data['data']
        except KeyError:
            log_data['body'] = '<NO BODY>'

        return log_data

    @coroutine
    def before_request(self, endpoint_desc, session, request_params):
        log_data = yield from self._prepare_on_request_log_record(endpoint_desc, session, request_params)
        self.logger.log(self.level, "Sending request", extra=log_data)

    @coroutine
    def on_response(self, endpoint_desc, session, request_params, response):
        log_data = yield from self._prepare_response_log_record(endpoint_desc, session, request_params, response)
        self.logger.log(self.level, "Response received", extra=log_data)


class OuterLogger(BaseLogger):

    @coroutine
    def _prepare_prepare_payload_log_record(self, endpoint_desc, session, request_params, payload):
        log_data = yield from self._prepare_request_log_record(endpoint_desc, session, request_params)
        if payload is not None:
            log_data['body'] = self._prepare_body(str(payload))
        else:
            log_data['body'] = '<NO BODY>'

        return log_data

    @coroutine
    def prepare_payload(self, endpoint_desc, session, request_params, payload):
        log_data = yield from self._prepare_prepare_payload_log_record(endpoint_desc, session,
                                                                       request_params, payload)
        self.logger.log(self.level, "Sending request", extra=log_data)

    @coroutine
    def on_parsed_response(self, endpoint_desc, session, request_params, response):
        log_data = yield from self._prepare_response_log_record(endpoint_desc, session, request_params, response)
        self.logger.log(self.level, "Response received", extra=log_data)
